<?php

namespace Drupal\lionbridge_translation_provider\Plugin\tmgmt\Translator;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\lionbridge_translation_provider\LionbridgeConnector;
use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt\TMGMTException;
use Drupal\tmgmt\TranslatorInterface;
use Drupal\tmgmt\TranslatorPluginBase;
use Drupal\tmgmt\Translator\AvailableResult;

/**
 * Lionbridge provider.
 *
 * @TranslatorPlugin(
 *   id = "lionbridge",
 *   label = @Translation("Lionbridge translator"),
 *   description = @Translation("Lionbridge translator service."),
 *   ui = "Drupal\lionbridge_translation_provider\LionbridgeTranslatorUi",
 *   default_settings = {
 *     "po_number" = "123456"
 *   },
 * )
 */
class LionbridgeTranslator extends TranslatorPluginBase {

  /**
   * Translation keys.
   *
   * @var array
   */
  protected $translationKeys;

  /**
   * {@inheritdoc}
   */
  public function requestTranslation(JobInterface $job) {
    if ($job->isUnprocessed()) {
      try {
        $this->submitJob($job);

        if (!$job->isRejected()) {
          $job->submitted(t('Job has been submitted.'));
        }
      }
      catch (TMGMTException $e) {
        watchdog_exception('lionbridge_translation_provider', $e);
        $job->rejected('Job has been rejected with following error: @error',
          array('@error' => $e->getMessage()), 'error');
      }
    }

    if ($job->isActive()) {
      try {
        $this->fetchJob($job);
      }
      catch (TMGMTException $e) {
        watchdog_exception('lionbridge_translation_provider', $e);
        $job->addMessage('Translation could not be completed: @error',
          array('@error' => $e->getMessage()), 'error');
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function checkAvailable(TranslatorInterface $translator) {
    if ($translator->getSetting('access_key_id') && $translator->getSetting('access_key')) {
      return AvailableResult::yes();
    }

    return AvailableResult::no(t('Access key ID and access key are not set.'));
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedRemoteLanguages(TranslatorInterface $translator) {
    if (!$translator->checkAvailable()->getSuccess()) {
      return parent::getSupportedRemoteLanguages($translator);
    }

    $api_client = new LionbridgeConnector($translator);
    $locales = $api_client->listLocales();
    $languages = [];
    foreach ($locales['Locale'] as $language) {
      $languages[$language['Code']] = $language['Code'];
    }

    return $languages;
  }

  /**
   * Submits a job to Lionbridge translation service.
   *
   * This does the work of actually submitting the translation job to the
   * Lionbridge translation service. There are 3 major parts to submitting a
   * translations job. Each part has been broken out into an individual
   * function to make the code more readable.
   *
   * 1. Create a JSON encoded file that contains all the translatable content
   *    and send it to Lionbridge. If this is successful, an array containing
   *    the details of the uploaded file is returned.
   *
   * 2. Create a project using the settings from the job settings page and the
   *    AssetID of the array returned from the successful file upload. If there
   *    were more than one file uploaded, multiple file asset ID's could be
   *    used in a project, however, in this model there is only 1 asset ID for
   *    a given project.
   *
   * 3. Generate a quote for the translation job using the Project ID from the
   *    array returned from a successful project creation request.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   The job object.
   */
  public function submitJob(JobInterface $job) {
    $translator = $job->getTranslator();
    $api_client = new LionbridgeConnector($translator);
    $items = $job->getItems();
    $file_assets = [];
    $source_language = $translator->mapToRemoteLanguage($job->getSourceLangcode());
    $target_language = $translator->mapToRemoteLanguage($job->getTargetLangcode());

    // This is the name of the JSON file that will be sent to Lionbridge.
    $file_name = $this->generateFileName($job->label(), $source_language, $target_language);

    if ($translator->getSetting('notification_url')) {
      $project_complete_url = $translator->getSetting('notification_url');
    }
    else {
      $project_complete_url = Url::fromRoute(
        'lionbridge_translation_provider.lionbridge_translation_provider_project_complete_callback',
        [],
        [
          'query' => ['secret' => $job->getSetting('secret')],
          'absolute' => TRUE,
        ]
      )->toString();
    }

    $file_asset = $this->generateFile($file_name, $source_language, $job->getData(), $api_client);

    if (isset($file_asset['Errors']['Error'])) {
      throw new TMGMTException($file_asset['Errors']['Error']['DetailedMessage']);
    }

    $file_assets[] = $file_asset['AssetID'];

    // Create a new project with the file asset, send it to Lionbridge. This
    // will create a new project and return the project XML in the form of an
    // array.
    $project_data = [
      'project_title' => $job->getSetting('project_title') . ' to ' . $target_language,
      'service_id' => $translator->getSetting('service'),
      'source_language' => $source_language,
      'target_languages' => [$target_language],
      'file_assets' => $file_assets,
    ];

    $project = $this->addProject($project_data, $api_client);

    if (isset($project['Errors']['Error']) && !empty($project['Errors']['Error'])) {
      throw new TMGMTException($project['Errors']['Error']['DetailedMessage']);
    }

    // Get a quote for the project.
    $quote = $this->generateQuote($project['ProjectID'], $project_complete_url, $api_client);

    if (isset($quote['Errors']) && !empty($quote['Errors'])) {
      throw new TMGMTException($quote['Errors']['Error']['DetailedMessage']);
    }

    // Set the job reference to the Quote ID.
    $job->reference = $quote['QuoteID'];

    // Add project ID to `tmgmt_remote` table.
    foreach ($this->translationKeys as $translation_key) {
      $items[$translation_key['tjiid']]->addRemoteMapping(
        $translation_key['data_item_key'],
        $project['ProjectID']
      );
    }
  }

  /**
   * Gets status of job items from Lionbridge and downloads translated items.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   The translation job object.
   *
   * @todo This needs to be fixed when "Poll translation" is implemented.
   */
  public function fetchJob(JobInterface $job) {
    $translator = $job->getTranslator();
    $api_client = new LionbridgeConnector($translator);
    if ($job->getReference()) {
      $quote = $api_client->getQuote($job->getReference());
      if (!$quote || isset($quote['Error'])) {
        throw new TMGMTException('Could not get quote.');
      }

      $translation_complete = FALSE;

      switch ($quote['Status']) {
        case LionbridgeConnector::QUOTE_STATUS_ERROR:
          $job->addMessage('An error occurred with this quote.');
          $this->abortTranslation($job);
          break;

        case LionbridgeConnector::QUOTE_STATUS_PENDING:
          // If this is a pending quote, add a message and a link to approve.
          $job->addMessage(
            'Quote ready for approval: <a href=":url" target="_blank">View on Lionbridge</a>', array(
              ':url' => Url::fromUri($api_client->getEndpoint() . '/project/' . $quote['QuoteID'] . '/details')->toString(),
            )
          );
          break;

        case LionbridgeConnector::QUOTE_STATUS_AUTHORIZED:
          $job->addMessage('Translation in progress.');
          break;

        case LionbridgeConnector::QUOTE_STATUS_COMPLETE:
          // Get the file which contains translations for all job items.
          // Loop through the remotes, populate translated data.
          $asset_id = $quote['Projects']['Project']['Files']['File']['AssetID'];
          $data_json = $api_client->getFileTranslation($asset_id, $translator->mapToRemoteLanguage($job->getTargetLangcode()));
          $data = Json::decode($data_json);

          if (!$data) {
            $data_error_message = json_last_error_msg();
            throw new TMGMTException($data_error_message);
          }

          // Batch process the storing of the translated data.
          batch_set(_lionbridge_translation_provider_batch($job, $data));
          $translation_complete = TRUE;
          break;

      }
      return $translation_complete;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function abortTranslation(JobInterface $job) {
    $translator = $job->getTranslator();
    $api_client = new LionbridgeConnector($translator);
    $quote = $api_client->getQuote($job->getReference());

    // If the status is Pending reject the quote on Lionbridge.
    if ($quote && empty($quote['Errors']) && $quote['Status'] == LionbridgeConnector::QUOTE_STATUS_PENDING) {
      $api_client->rejectQuote($quote['QuoteID']);
    }

    $job->aborted();
    return TRUE;
  }

  /**
   * Sends a file to the Lionbridge translation service.
   */
  protected function addFile($translation, LionbridgeConnector $api_client) {
    return $api_client->addFile(
      $translation['source_language'],
      $translation['fileName'],
      $translation['content_type'],
      $translation['file_content']
    );
  }

  /**
   * Sends a create project request to the Lionbridge translation service.
   */
  protected function addProject($project_data, LionbridgeConnector $api_client) {
    return $api_client->addProject(
      $project_data['project_title'],
      $project_data['service_id'],
      $project_data['source_language'],
      $project_data['target_languages'],
      $project_data['file_assets']
    );
  }

  /**
   * Gets a quote from the Lionbridge translation service.
   */
  protected function generateQuote($project_id, $notification_url, LionbridgeConnector $api_client) {
    return $api_client->generateQuote($project_id, $notification_url);
  }

  /**
   * Create JSON encoded file.
   *
   * @param string $file_name
   *   The lionbridge connector.
   * @param string $source_language
   *   The source language.
   * @param mixed $data
   *   The data to be flattened.
   * @param LionbridgeConnector $api_client
   *   The API client.
   *
   * @return array
   *   The created file.
   */
  protected function generateFile($file_name, $source_language, $data, LionbridgeConnector $api_client) {
    $translatable_content = array();

    $data_flat = array_filter(\Drupal::service('tmgmt.data')->flatten($data), array(\Drupal::service('tmgmt.data'), 'filterData'));

    foreach ($data_flat as $key => $value) {
      list($tjiid, $data_item_key) = explode('][', $key, 2);

      // Preserve the original data.
      $this->translationKeys[] = array(
        'tjiid' => $tjiid,
        'data_item_key' => $data_item_key,
      );

      // This is the actual text sent to Lionbridge for translation.
      $translatable_content[$tjiid][$data_item_key] = str_replace(array('“', '”'), '\"', $value['#text']);
    }

    $file_content = json_encode($translatable_content, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE);

    $translation = array(
      'source_language' => $source_language,
      'fileName' => urlencode($file_name . '.json'),
      'content_type' => 'application/json',
      'file_content' => $file_content,
    );

    return $this->addFile($translation, $api_client);
  }

  /**
   * Create a file name.
   *
   * @param string $file_name
   *   The file name.
   * @param string $source_language
   *   The source language code.
   * @param string $target_language
   *   The target language code.
   *
   * @return string
   *   The generated file name.
   */
  protected function generateFileName($file_name, $source_language, $target_language) {
    $length = 0;
    $file_name_string = '';
    $string = strtolower($file_name);

    if (preg_match_all('/(\w+)/is', $string, $matches)) {
      foreach ($matches[0] as $match) {
        $length += strlen($match);

        if ($length < 45) {
          $file_name_string .= $match . '-';
        }
      }

      $file_name_string .= $source_language . '-' . $target_language;
      return $file_name_string;
    }

    // Something went wrong, just return $string.
    return $string;
  }

}
